from abc import ABCMeta, abstractmethod

BEFORE_REQUEST = 0x1
AFTER_RESPONSE = 0x2


class MiddleWareInterface(metaclass=ABCMeta):
    type = BEFORE_REQUEST
    # type = AFTER_RESPONSE

    @abstractmethod
    def apply_before(self, request, **kwargs):
        pass

    @abstractmethod
    def apply_after(self, response, **kwargs):
        pass


class MiddlewareManager:
    def __init__(self, middlewares=None):
        if not middlewares:
            self.middlewares = list()
        else:
            self.middlewares = middlewares

    def register(self, middlewares):
        self.middlewares.extend(middlewares)

    def apply_before(self, request):
        for middleware in self.middlewares:
            if middleware.type == BEFORE_REQUEST:
                middleware.apply_before(request)
        return request

    def apply_after(self, request, response):
        for middleware in self.middlewares:
            if middleware.type == AFTER_RESPONSE:
                middleware.apply_after(response, request=request)
        return response




from werkzeug.wrappers import Response
from jinja2 import Environment, FileSystemLoader


class TextPlainResponse(Response):
    mine_type = 'text/plain'

    def __init__(self, text):
        super().__init__(response=text, status=200, mimetype=self.mine_type)


class TextHTMLResponse(Response):
    mine_type = 'text/html'

    def __init__(self, context=None, template_path='', template_name=''):
        self.jinja = Environment(
            loader=FileSystemLoader(template_path),
            autoescape=True
        )
        template = self.jinja.get_template(template_name)
        rendered_text = template.render(context=context)
        super().__init__(response=rendered_text, mimetype=self.mine_type)



from werkzeug.exceptions import HTTPException
from werkzeug.serving import run_simple
from werkzeug.wrappers import Request

from JJCale.middlewares import MiddlewareManager
from JJCale.router import Router


class JJCale:
    def __init__(self):
        self.middleware_manager = MiddlewareManager()
        self.router = Router()

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        request = self.before_request(request)

        try:
            view_func, kwargs = self.router.route(request)
            response = view_func(request, **kwargs)
        except HTTPException as e:
            return e

        response = self.after_response(request, response)
        return response.__call__(environ, start_response)

    def before_request(self, request):
        return self.middleware_manager.apply_before(request)

    def after_response(self, request, response):
        return self.middleware_manager.apply_after(request, response)

    def run(self, host='localhost', port=23333, debug=False):
        run_simple(
            hostname=host,
            port=port,
            application=self.wsgi_app,
            use_reloader=debug,
            use_debugger=debug,
        )



from setuptools import setup, find_packages

setup(
    name='JJCale',
    version='0.0.3.1',
    description='a simple web framework',
    long_description='a simple web framework',
    url='https://github.com/lrhhhhhh/JJCale',
    author='lrhaoo',
    license='MIT',
    keywords=('web', 'framework'),
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'Werkzeug>=2.0.0',
        'jinja2>=3.0.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



from werkzeug.routing import Map, Rule
from werkzeug.exceptions import NotFound
from dataclasses import dataclass


@dataclass
class SubRouter:
    url: str
    view_func: callable
    endpoint: str

    def __iter__(self):
        return iter([self.url, self.view_func, self.endpoint])


class Router:
    def __init__(self):
        self.rules = dict()
        self.url_set = set()
        self.map = Map()

    def merge(self, url, sub_router, endpoint):
        sub_url, view_func, sub_endpoint = sub_router
        new_url = url + sub_url
        new_endpoint = endpoint + '.' + sub_endpoint,
        self.add(new_url, view_func, new_endpoint)

    def add(self, url, view_func, endpoint):
        if url in self.url_set:
            raise ValueError(f'url {url} already exists')
        else:
            self.url_set.add(url)

        if endpoint in self.rules:
            raise ValueError(f'endpoint {endpoint} already exists')

        self.map.add(
            Rule(url, endpoint=endpoint)
        )
        self.rules[endpoint] = {
            'view_func': view_func,
            'url': url
        }

    def register(self, rules):
        for rule in rules:
            url, x, endpoint = rule

            if isinstance(x, SubRouter):
                self.merge(url, x, endpoint)
            elif callable(x):
                self.add(url, x, endpoint)
            else:
                raise TypeError('Expect router or view_func')

    def route(self, request):
        map_adapter = self.map.bind_to_environ(request.environ)
        try:
            endpoint, kwargs = map_adapter.match()
            view_func = self.rules.get(endpoint)['view_func']
            return view_func, kwargs
        except NotFound as e:
            raise NotFound(f'path {request.path} not found')

    def reverse(self, endpoint):
        res = self.rules.get(endpoint)
        if res:
            return res['url']
        else:
            return None


if __name__ == '__main__':
    routes = [
        ('/index/something', SubRouter(url='/people', view_func=lambda: 'fake view func1', endpoint='people'), 'index'),
        ('/blog', SubRouter(url='/article1', view_func=lambda: 'fake view func2', endpoint='article'), 'blog'),
        ('/about', lambda: 'fake view func3', 'about'),
    ]
    router = Router()
    router.register(routes)
    print(router.rules)
    print(router.map)
